"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
const ts = require("typescript");
const CommandLineParser_1 = require("./CommandLineParser");
const Transpiler_51_1 = require("./targets/Transpiler.51");
const Transpiler_52_1 = require("./targets/Transpiler.52");
const Transpiler_53_1 = require("./targets/Transpiler.53");
const Transpiler_JIT_1 = require("./targets/Transpiler.JIT");
const Transpiler_1 = require("./Transpiler");
function compile(argv) {
    const commandLine = CommandLineParser_1.parseCommandLine(argv);
    compileFilesWithOptions(commandLine.fileNames, commandLine.options);
}
exports.compile = compile;
function compileFilesWithOptions(fileNames, options) {
    if (!options.luaTarget) {
        options.luaTarget = Transpiler_1.LuaTarget.LuaJIT;
    }
    const program = ts.createProgram(fileNames, options);
    const checker = program.getTypeChecker();
    // Get all diagnostics, ignore unsupported extension
    const diagnostics = ts.getPreEmitDiagnostics(program).filter(diag => diag.code !== 6054);
    diagnostics.forEach(diagnostic => {
        if (diagnostic.file) {
            const { line, character } = diagnostic.file.getLineAndCharacterOfPosition(diagnostic.start);
            const message = ts.flattenDiagnosticMessageText(diagnostic.messageText, "\n");
            console.log(`${diagnostic.file.fileName} (${line + 1},${character + 1}): ${message}`);
        }
        else {
            console.log(`${ts.flattenDiagnosticMessageText(diagnostic.messageText, "\n")}`);
        }
    });
    // If there are errors dont emit
    if (diagnostics.filter(diag => diag.category === ts.DiagnosticCategory.Error).length > 0) {
        console.log("Stopping compilation process because of errors.");
        process.exit(1);
    }
    program.getSourceFiles().forEach(sourceFile => {
        if (!sourceFile.isDeclarationFile) {
            try {
                const rootDir = options.rootDir;
                // Transpile AST
                const lua = createTranspiler(checker, options, sourceFile).transpileSourceFile();
                let outPath = sourceFile.fileName;
                if (options.outDir !== options.rootDir) {
                    const relativeSourcePath = path.resolve(sourceFile.fileName)
                        .replace(path.resolve(rootDir), "");
                    outPath = path.join(options.outDir, relativeSourcePath);
                }
                // change extension or rename to outFile
                if (options.outFile) {
                    if (path.isAbsolute(options.outFile)) {
                        outPath = options.outFile;
                    }
                    else {
                        // append to workingDir or outDir
                        outPath = path.resolve(options.outDir, options.outFile);
                    }
                }
                else {
                    const fileNameLua = path.basename(outPath, path.extname(outPath)) + ".lua";
                    outPath = path.join(path.dirname(outPath), fileNameLua);
                }
                // Write output
                ts.sys.writeFile(outPath, lua);
            }
            catch (exception) {
                /* istanbul ignore else: Testing else part would require to add a bug/exception to our code */
                if (exception.node) {
                    const pos = ts.getLineAndCharacterOfPosition(sourceFile, exception.node.pos);
                    // Graciously handle transpilation errors
                    console.error("Encountered error parsing file: " + exception.message);
                    console.error(sourceFile.fileName + " line: " + (1 + pos.line) + " column: " + pos.character + "\n" +
                        exception.stack);
                    process.exit(1);
                }
                else {
                    throw exception;
                }
            }
        }
    });
    // Copy lualib to target dir
    if (options.luaLibImport === Transpiler_1.LuaLibImportKind.Require || options.luaLibImport === Transpiler_1.LuaLibImportKind.Always) {
        fs.copyFileSync(path.resolve(__dirname, "../dist/lualib/lualib_bundle.lua"), path.join(options.outDir, "lualib_bundle.lua"));
    }
}
exports.compileFilesWithOptions = compileFilesWithOptions;
function createTranspiler(checker, options, sourceFile) {
    let luaTargetTranspiler;
    switch (options.luaTarget) {
        case Transpiler_1.LuaTarget.LuaJIT:
            luaTargetTranspiler = new Transpiler_JIT_1.LuaTranspilerJIT(checker, options, sourceFile);
            break;
        case Transpiler_1.LuaTarget.Lua51:
            luaTargetTranspiler = new Transpiler_51_1.LuaTranspiler51(checker, options, sourceFile);
            break;
        case Transpiler_1.LuaTarget.Lua52:
            luaTargetTranspiler = new Transpiler_52_1.LuaTranspiler52(checker, options, sourceFile);
            break;
        case Transpiler_1.LuaTarget.Lua53:
            luaTargetTranspiler = new Transpiler_53_1.LuaTranspiler53(checker, options, sourceFile);
            break;
        default:
            // should not happen
            throw Error("No luaTarget Specified please ensure a target is set!");
    }
    return luaTargetTranspiler;
}
exports.createTranspiler = createTranspiler;
//# sourceMappingURL=Compiler.js.map