"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
class TSHelper {
    // Reverse lookup of enum key by value
    static enumName(needle, haystack) {
        for (const name in haystack) {
            if (haystack[name] === needle) {
                return name;
            }
        }
        return "unknown";
    }
    // Breaks down a mask into all flag names.
    static enumNames(mask, haystack) {
        const result = [];
        for (const name in haystack) {
            if ((mask & haystack[name]) !== 0 && mask >= haystack[name]) {
                result.push(name);
            }
        }
        return result;
    }
    static containsStatement(statements, kind) {
        return statements.some(statement => statement.kind === kind);
    }
    static getExtendedType(node, checker) {
        if (node.heritageClauses) {
            for (const clause of node.heritageClauses) {
                if (clause.token === ts.SyntaxKind.ExtendsKeyword) {
                    const superType = checker.getTypeAtLocation(clause.types[0]);
                    if (!this.isPureAbstractClass(superType, checker)) {
                        return superType;
                    }
                }
            }
        }
        return undefined;
    }
    static isFileModule(sourceFile) {
        if (sourceFile) {
            // Vanilla ts flags files as external module if they have an import or
            // export statement, we only check for export statements
            return sourceFile.statements.some(statement => (ts.getCombinedModifierFlags(statement) & ts.ModifierFlags.Export) !== 0
                || statement.kind === ts.SyntaxKind.ExportAssignment
                || statement.kind === ts.SyntaxKind.ExportDeclaration);
        }
        return false;
    }
    static isInDestructingAssignment(node) {
        return node.parent && ts.isVariableDeclaration(node.parent) && ts.isArrayBindingPattern(node.parent.name);
    }
    static isStringType(type) {
        return (type.flags & ts.TypeFlags.String) !== 0
            || (type.flags & ts.TypeFlags.StringLike) !== 0
            || (type.flags & ts.TypeFlags.StringLiteral) !== 0;
    }
    static isArrayType(type, checker) {
        const typeNode = checker.typeToTypeNode(type);
        return typeNode && (typeNode.kind === ts.SyntaxKind.ArrayType || typeNode.kind === ts.SyntaxKind.TupleType);
    }
    static isCompileMembersOnlyEnum(type, checker) {
        return type.symbol
            && ((type.symbol.flags & ts.SymbolFlags.Enum) !== 0)
            && type.symbol.getDocumentationComment(checker)[0] !== undefined
            && this.hasCustomDecorator(type, checker, "!CompileMembersOnly");
    }
    static isPureAbstractClass(type, checker) {
        return type.symbol
            && ((type.symbol.flags & ts.SymbolFlags.Class) !== 0)
            && this.hasCustomDecorator(type, checker, "!PureAbstract");
    }
    static isExtensionClass(type, checker) {
        return type.symbol
            && ((type.symbol.flags & ts.SymbolFlags.Class) !== 0)
            && this.hasCustomDecorator(type, checker, "!Extension");
    }
    static isPhantom(type, checker) {
        return type.symbol
            && ((type.symbol.flags & ts.SymbolFlags.Namespace) !== 0)
            && this.hasCustomDecorator(type, checker, "!Phantom");
    }
    static isTupleReturnCall(node, checker) {
        if (ts.isCallExpression(node)) {
            const type = checker.getTypeAtLocation(node.expression);
            return this.isTupleReturnFunction(type, checker);
        }
        else {
            return false;
        }
    }
    static isTupleReturnFunction(type, checker) {
        return type.symbol
            && ((type.symbol.flags & ts.SymbolFlags.Function) !== 0
                || (type.symbol.flags & ts.SymbolFlags.Method) !== 0)
            && this.hasCustomDecorator(type, checker, "!TupleReturn");
    }
    static hasCustomDecorator(type, checker, decorator) {
        if (type.symbol) {
            const comments = type.symbol.getDocumentationComment(checker);
            const decorators = comments.filter(comment => comment.kind === "text")
                .map(comment => comment.text.trim())
                .filter(comment => comment[0] === "!");
            return decorators.indexOf(decorator) > -1;
        }
        return false;
    }
    // Search up until finding a node satisfying the callback
    static findFirstNodeAbove(node, callback) {
        let current = node;
        while (current.parent) {
            if (callback(current.parent)) {
                return current.parent;
            }
            else {
                current = current.parent;
            }
        }
        return null;
    }
    static hasGetAccessor(node, checker) {
        if (ts.isPropertyAccessExpression(node)) {
            const name = node.name.escapedText;
            const type = checker.getTypeAtLocation(node.expression);
            if (type && type.symbol && type.symbol.members) {
                const field = type.symbol.members.get(name);
                return field && (field.flags & ts.SymbolFlags.GetAccessor) !== 0;
            }
        }
        return false;
    }
    static hasSetAccessor(node, checker) {
        if (ts.isPropertyAccessExpression(node)) {
            const name = node.name.escapedText;
            const type = checker.getTypeAtLocation(node.expression);
            if (type && type.symbol && type.symbol.members) {
                const field = type.symbol.members.get(name);
                return field && (field.flags & ts.SymbolFlags.SetAccessor) !== 0;
            }
        }
        return false;
    }
    static isBinaryAssignmentToken(token) {
        switch (token) {
            case ts.SyntaxKind.BarEqualsToken:
                return [true, ts.SyntaxKind.BarToken];
            case ts.SyntaxKind.PlusEqualsToken:
                return [true, ts.SyntaxKind.PlusToken];
            case ts.SyntaxKind.CaretEqualsToken:
                return [true, ts.SyntaxKind.CaretToken];
            case ts.SyntaxKind.MinusEqualsToken:
                return [true, ts.SyntaxKind.MinusToken];
            case ts.SyntaxKind.SlashEqualsToken:
                return [true, ts.SyntaxKind.SlashToken];
            case ts.SyntaxKind.PercentEqualsToken:
                return [true, ts.SyntaxKind.PercentToken];
            case ts.SyntaxKind.AsteriskEqualsToken:
                return [true, ts.SyntaxKind.AsteriskToken];
            case ts.SyntaxKind.AmpersandEqualsToken:
                return [true, ts.SyntaxKind.AmpersandToken];
            case ts.SyntaxKind.AsteriskAsteriskEqualsToken:
                return [true, ts.SyntaxKind.AsteriskAsteriskToken];
            case ts.SyntaxKind.LessThanLessThanEqualsToken:
                return [true, ts.SyntaxKind.LessThanLessThanToken];
            case ts.SyntaxKind.GreaterThanGreaterThanEqualsToken:
                return [true, ts.SyntaxKind.GreaterThanGreaterThanToken];
            case ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken:
                return [true, ts.SyntaxKind.GreaterThanGreaterThanGreaterThanToken];
        }
        return [false, null];
    }
    static isExpressionStatement(node) {
        return node.parent === undefined
            || ts.isExpressionStatement(node.parent)
            || ts.isForStatement(node.parent);
    }
}
exports.TSHelper = TSHelper;
//# sourceMappingURL=TSHelper.js.map